"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const amqplib_1 = require("amqplib");
const MessageBusEventEmitter_1 = require("../messageBus/MessageBusEventEmitter");
const __1 = require("..");
const MAX_RETRIES = 14;
const DEFAULT_MS_STEP = 1000;
class RabbitMQConnectionFacade {
    constructor(serviceName, appName, delayRetry, filterTrafic, preFetch) {
        this.exchange = serviceName;
        this.deadLetterExchange = `${this.exchange}_dlx`;
        this.queue = `${serviceName}_queue`;
        this.retryQueue = `${this.queue}_retry`;
        this.appName = appName;
        this.delayRetry = delayRetry;
        this.messagePool = new Map();
        this.filterTrafic = filterTrafic;
        this.preFetch = preFetch;
    }
    async subscribeAll(messageBindings) {
        try {
            for (const [key] of messageBindings) {
                await this.subscribe(key);
            }
            return true;
        }
        catch (err) {
            return false;
        }
    }
    async disconnect() {
        if (this.connection) {
            try {
                this.channel.removeAllListeners();
                await this.channel.close();
                await this.connection.close();
                this.connection = undefined;
                return true;
            }
            catch (e) {
                throw e;
            }
        }
        else {
            return false;
        }
    }
    async connect(hostname, username, password) {
        try {
            this.connection = await amqplib_1.connect({
                hostname,
                password,
                port: 5672,
                protocol: 'amqp',
                username,
            });
            this.connection.on('error', () => { });
            this.channel = await this.connection.createChannel();
            this.channel.on('error', error => {
                console.error(error);
            });
            this.channel.on('close', () => {
                console.log('rabbitmq channel closed');
            });
            await this.setUp();
        }
        catch (e) {
            throw e;
        }
        const emitter = MessageBusEventEmitter_1.MessageEmitter.getMessageEmitter();
        emitter.on(MessageBusEventEmitter_1.MessageEmitter.PROCESS_SUCCESS, (resultingMessages) => {
            if (resultingMessages) {
                for (const message of resultingMessages) {
                    try {
                        this.publish(message);
                    }
                    catch (e) {
                        emitter.emit('error', e);
                    }
                }
            }
        });
        try {
            this.channel.prefetch(this.preFetch);
            await this.channel.consume(`${this.queue}`, (msg) => {
                if (!this.filterTrafic || msg.properties.appId === this.appName) {
                    if (msg.properties.messageId) {
                        const descriptor = msg.fields.routingKey;
                        this.messagePool.set(msg.properties.messageId, msg);
                        this.dispatchMessage(descriptor, msg);
                    }
                    else {
                        this.channel.ack(msg);
                    }
                }
                else {
                    this.channel.ack(msg);
                }
            });
            return this.connection;
        }
        catch (e) {
            throw e;
        }
    }
    publish(message, alternativeExchange = null) {
        const exchange = alternativeExchange ? alternativeExchange : this.appName;
        try {
            return this.channel.publish(exchange, message.getDescriptor(), Buffer.from(message.getSerializedContent()), {
                appId: this.appName,
                headers: { parentMessage: message.getParentMessage() },
                messageId: message.getUUID(),
                persistent: true,
                timestamp: new Date().getTime(),
            });
        }
        catch (e) {
            throw e;
        }
    }
    async subscribe(descriptor) {
        try {
            await this.channel.bindQueue(this.queue, this.exchange, descriptor);
            return descriptor;
        }
        catch (e) {
            throw e;
        }
    }
    async dispatchMessage(descriptor, msg) {
        const handlers = __1.default.handlerRegistry;
        if (Reflect.hasMetadata('messageBindings', __1.default.prototype.constructor)) {
            const bindings = Reflect.getMetadata('messageBindings', __1.default.prototype.constructor);
            const handlerClassName = bindings.get(descriptor);
            if (handlerClassName) {
                if (handlers.has(handlerClassName)) {
                    const handler = handlers.get(handlerClassName);
                    try {
                        const MercuryMessage = new __1.JSONMessage(descriptor, msg.content, msg.properties.messageId, msg.properties.timestamp, msg.properties.headers.parentMessage);
                        const result = await handler.handle(MercuryMessage);
                        const resultMessages = Array.isArray(result) && result.every(r => r instanceof __1.Message)
                            ? result
                            : result instanceof __1.Message
                                ? [result]
                                : null;
                        this.channel.ack(this.messagePool.get(msg.properties.messageId));
                        if (resultMessages) {
                            for (const message of resultMessages) {
                                this.publish(message);
                            }
                        }
                        this.messagePool.delete(msg.properties.messageId);
                    }
                    catch (e) {
                        this.messagePool.delete(msg.properties.messageId);
                        if (!msg.properties.headers['x-death'] ||
                            (msg.properties.headers['x-death'] &&
                                msg.properties.headers['x-death'][0].count < MAX_RETRIES)) {
                            this.channel.nack(msg, false, false);
                        }
                        else {
                            this.channel.ack(msg);
                        }
                    }
                }
                else {
                    this.channel.ack(msg);
                }
            }
        }
    }
    async setUp() {
        /* Creating Exchanges and queues */
        try {
            await this.channel.assertExchange(this.appName, 'fanout', {
                autoDelete: false,
                durable: true,
            });
            await this.channel.assertExchange(this.exchange, 'direct', {
                autoDelete: false,
                durable: true,
            });
            await this.channel.assertExchange(this.deadLetterExchange, 'fanout', {
                autoDelete: false,
                durable: true,
            });
            await this.channel.assertQueue(this.queue, {
                autoDelete: false,
                deadLetterExchange: this.deadLetterExchange,
                durable: true,
            });
            await this.channel.assertQueue(this.retryQueue, {
                arguments: {
                    'x-message-ttl': this.delayRetry * DEFAULT_MS_STEP,
                },
                autoDelete: false,
                deadLetterExchange: this.exchange,
                durable: true,
            });
            /* Creating the basic bindings */
            await this.channel.bindExchange(this.exchange, this.appName, '');
            await this.channel.bindQueue(this.retryQueue, this.deadLetterExchange, '');
        }
        catch (e) {
            throw e;
        }
        return true;
    }
}
exports.RabbitMQConnectionFacade = RabbitMQConnectionFacade;
//# sourceMappingURL=RabbitMQConnectionFacade.js.map